"""
IRC Client library for Python 2.2

07/05/2003 - 19:47:13
GNU-GPL v2, (C)2003 Romain GUY and Helix Team <romain.guy@jext.org>
"""

# ========================================================================================
# IMPORT STATEMENTS
# ========================================================================================

import re                 # regex
import socket             # network sockets
import string             # string manipulations
import sys                # system access
import threading          # threads
import time               # time manipulation
import types              # data types
import Queue              # queue list

# ========================================================================================
# MODULE LEVEL VARIABLES
# ========================================================================================

irclib_debug = 0

# ========================================================================================
# IRCBOT CLASS
# ========================================================================================

class IRCBot:
  queueProcessing = { }

  def __init__(self, remoteHost, port = 6667, realName = "irclib", userName = "irclib", server = "localhost", chans = "irclib"):
    self.chans           = chans
    self.connection      = IRCClient(self, remoteHost, port, realName, userName, server)
    self.queueProcessing = { "PING"            : self.processPing,
                             "ERROR"           : self.processError,
                             "WELCOME"         : self.processWelcomeMessage,
                             "LOGGED"          : self.processLogged
                           }
    self._stopEvent = threading.Event()
    self._sleepPeriod = 1.0

  def addProcess(self, message, function):
    self.queueProcessing[message] = function

  def connect(self):
    return self.connection.connect()

  def getConnection(self):
    return self.connection

  def processIRCMessage(self):
    while self.ircQueue.qsize():
      try:
        args    = self.ircQueue.get(0)
        msgType = args[0]

        if msgType in self.queueProcessing:
          self.queueProcessing[msgType](args[1:])
      except Queue.Empty:
        pass

  def processError(self, args):
    pass

  def processLogged(self, args):
    self.connection.join(self.chans.split(","))

  def processPing(self, args):
    if irclib_debug: print "PING ? PONG !"
    text = args[0]
    self.connection.send('PONG ' + text + '\r\n')

  def processWelcomeMessage(self, args):
    line = args[0]
    if line.find('PING') == 0:
      match = re.search("^PING :(\d+)", line)
      self.connection.send('PONG ' + match.group(1) + '\r\n')

  def run(self):
    if self.connect():
      while self.connection.running:
        self._stopEvent.wait(self._sleepPeriod)

  def send(self, dest, msg):
    self.connection.send("PRIVMSG %s :%s\r\n" % (dest, msg))

# ========================================================================================
# IRCCLIENT CLASS
# ========================================================================================

class IRCClient:
  def __init__(self, gui, remoteHost, port = 6667, realName = "irclib", userName = "irclib", server = "localhost"):
    self.nick          = 'irclib'
    self.realName      = realName
    self.userName      = userName
    self.server        = server
    self.host          = 'localhost'
    self.remoteHost    = remoteHost
    self.port          = port
    self.ircConnection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.waitingFor    = 0
    self.queue         = Queue.Queue()
    self.gui           = gui
    self.gui.ircQueue  = self.queue

  def __messagesPump(self):
    if not self.gui.processIRCMessage:
      return

    self._sleeper = threading.Event()
    while self.running:
      self.gui.processIRCMessage()
      self._sleeper.wait(0.3)

  def endApplication(self):
    self.running = 0
  def getNick(self):
    return self.nick
  def getRealName(self):
    return self.realName
  def getUserName(self):
    return self.userName
  def getServer(self):
    return self.server
  def getHost(self):
    return self.host
  def getRemoteHost(self):
    return self.remoteHost
  def getPort(self):
    return self.port

  def setNick(self, nick):
    self.nick = nick
  def setRealName(self, realName):
    self.realName = realName
  def setUserName(userName):
    self.userName = userName
  def setServer(self, server):
    self.server = server
  def setHost(self, hote):
    self.hote = hote
  def setRemoteHost(self, remoteHost):
    self.remoteHost = remoteHost
  def setPort(self, port):
    self.port = port

  def startIdentdServer(self):
    self.identdServer = threading.Thread(target = self.__identdServer)
    self.identdServer.setDaemon(1)
    self.identdServer.start()

  def __identdServer(self):
    identd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    identd.bind(('', 113))
    identd.listen(1)
    irc, irchost = identd.accept()
    self.queue.put(("IDENTD-REQUEST", irchost))
    match = re.search('(\d+)', irc.recv(512))
    response = ""
    if match is not None:
      response = match.group(1) + ", " + str(self.getPort()) + ' : USERID : UNIX : ' + self.getUserName() + '\r\n'
    else:
      str(self.getPort()) + ' : USERID : UNIX : ' + self.getUserName() + '\r\n'
    irc.send(response)
    self.queue.put(("IDENTD-RESPONSE", response))
    irc.close()
    identd.close()
      
  def testConnection(self):
    self.connectionTest = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
      self.connectionTest.connect((self.getRemoteHost(), self.getPort()))
      self.connectionTest.close()
      return 1
    except:
      self.connectionTest.close()
      return 0

  def connect(self):
    if not self.testConnection(): return 0
    
    self.startIdentdServer()
      
    self.ircConnection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.ircConnection.connect((self.getRemoteHost(), self.getPort()))

    try:
      ircCommand = "NICK "
      ircCommand = ircCommand + self.getNick()+'\r\n'
      self.ircConnection.send(ircCommand)
      ircCommand = "USER "
      ircCommand = ircCommand + self.getUserName() +' ' + self.getRemoteHost() + ' '
      ircCommand = ircCommand + self.getServer() + ' ' + ':' + self.getRealName()     + '\r\n'
      self.ircConnection.send(ircCommand)
    except:
      self.waitThread = threading.Thread(target = self.__wait)
      self.waitThread.setDaemon(1)
      self.waitThread.start()
      return 0

    self.threadPrincipalLoop()

    return 1

  def __wait(self):
    self.waitingFor   = 1
    self.identdServer = None
    if irclib_debug: print "*** start waiting"
    time.sleep(30)
    if irclib_debug: print "*** end waiting"
    self.waitingFor = 0
    self.queue.put(("END-WAITING", ()))

  def isWaitingFor(self):
    return self.waitingFor

  def threadPrincipalLoop(self):
    self.running      = 1
    self.threadBoucle = threading.Thread(target = self.__principalLoop)
    self.threadBoucle.setDaemon(1)
    self.threadBoucle.start()

    self.timer = threading.Thread(target = self.__messagesPump)
    self.timer.setDaemon(1)
    self.timer.start()

  def closeConnection(self):
    self.endApplication()
    self.ircConnection.close()

  def getMsgFromBuffer(self, buffer):
    if buffer == "":
      return ([""],"")

    full = 0

    if buffer[-1] == '\n' and buffer[-2] == '\r':full = 1

    if full :
      l = string.split(buffer,'\r\n')
      buffer = ''
      return (l, buffer)
    else:

      buffer = buffer +'\r\n'
      l = string.split(buffer,'\r\n')
      buffer = l[-2]
      return (l[:-2], buffer)

  def send(self, text):
    try:
      self.ircConnection.send(text)
    except:
      self.queue.put(("ERROR", ))

  def oper(self, psw):
    ircCommand = "OPER " + self.getNick() + " " + psw + "\r\n"
    self.send(ircCommand)

  def who(self, name, listOp = ""):
    ircCommand = "WHO " + name + " " + listOp + "\r\n"
    self.send(ircCommand)

  def whoIs(self, nick, server=""):
    ircCommand = "WHOIS " + server + " " + nick + "\r\n"
    self.send(ircCommand)

  def whoWas(self, nick, counter = "", server = ""):
    ircCommand = "WHOWAS " + nick + " " + counter + " " + server + "\r\n"
    self.send(ircCommand)

  def kill(self, nick, comment = ""):
    ircCommand = "KILL " + nick + " " + comment + "\r\n"
    self.send(ircCommand)

  def quit(self, message = ""):
    ircCommand = "QUIT :" + message + "\r\n"
    self.send(ircCommand)

  def away(self, message = ""):
    ircCommand = "AWAY :" + message + "\r\n"
    self.send(ircCommand)
    
  def back(self):
    ircCommand = "AWAY\r\n"
    self.send(ircCommand) 

  def join(self, chanList, keyList = [""]):
    ircCommand = "JOIN "
    i= 0
    nbi = len(chanList)
    for truc in chanList[:-1]:
      ircCommand = ircCommand + truc + ","
    ircCommand = ircCommand + chanList[-1] + " "

    for truc in keyList[:-1]:
      ircCommand = ircCommand + truc + ","
    ircCommand = ircCommand + keyList[-1]

    ircCommand = ircCommand +"\r\n"
    self.send(ircCommand)

  def part(self, chanList):
    ircCommand = "PART "

    for truc in chanList[:-1]:
      ircCommand = ircCommand + truc + ","
    ircCommand = ircCommand + chanList[-1] + " "

    ircCommand = ircCommand + "\r\n"
    self.send(ircCommand)

  def modeChan(self, chan, modeList, users, limit = "" , mask = ""):
    ircCommand = "MODE " + chan + " "
    for truc in modeList:
      ircCommand = ircCommand + truc

    ircCommand = ircCommand + " " + limit + " "
    for truc in users:
      ircCommand = ircCommand + " " + truc + " "

    ircCommand = ircCommand + mask + "\r\n"
    self.send(ircCommand)

  def modeUser(self, user, modeList):
    ircCommand = "MODE " + user + " "
    for truc in modeList:
      ircCommand = ircCommand + truc

    ircCommand = ircCommand + "\r\n"
    self.send(ircCommand)

  def topic(self, chan, newTopic=""):
    ircCommand = "TOPIC " + chan + " "

    if newTopic != "":
      ircCommand = ircCommand + ":" + newTopic+ " "

    ircCommand = ircCommand + "\r\n"
    self.send(ircCommand)

  def names(self, chanList = [""]):
    ircCommand = "NAMES "

    for truc in chanList[:-1]:
      ircCommand = ircCommand + truc + ","

    ircCommand = ircCommand + chanList[-1] + "\r\n"
    self.send(ircCommand)

  def list(self):
    ircCommand = "LIST" + "\r\n"
    self.send(ircCommand)

  def privMsg(self, targetList, message):
    ircCommand = "PRIVMSG "
    for truc in targetList[:-1]:
      ircCommand = ircCommand + truc + ","

    ircCommand = ircCommand + targetList[-1] + " :" + message + "\r\n"
    self.send(ircCommand)

  def notice(self, targetList, message):
    ircCommand = "NOTICE "
    for truc in targetList[:-1]:
      ircCommand = ircCommand + truc + ","

    ircCommand = ircCommand + targetList[-1] + " :" + message + "\r\n"
    self.send(ircCommand)

  def invite(self, user, chan):
    ircCommand = "INVITE " + user + " " + chan + "\r\n"
    self.send(ircCommand)

  def kick(self, chan, user, comment=""):
    ircCommand = "KICK " + chan + " " + user
    if comment != "":
      ircCommand = ircCommand + " :" + comment

    ircCommand = ircCommand + "\r\n"
    self.send(ircCommand)

  def version(self, server = ""):
    ircCommand = "VERSION "
    if server != "":
      ircCommand = ircCommand + server

    ircCommand = ircCommand + "\r\n"
    self.send(ircCommand)

  def ctcpVersion(self, target):
    self.send("PRIVMSG " + target + " :" + chr(1) + "VERSION" + chr(1) + '\r\n')

  def ctcpTime(self, target):
    self.send("PRIVMSG " + target + " :" + chr(1) + "TIME" + chr(1) + "\r\n")

  def ctcpPing(self, target):
    self.send("PRIVMSG " + target + " :" + chr(1) + "PING " + str(int(time.time())) + chr(1) + "\r\n")

  def ctcpSound(self, target, file):
    self.send("PRIVMSG " + target + " :" + chr(1) + "SOUND " + file + chr(1) + "\r\n")

  def ctcpImage(self, target, url):
    self.send("PRIVMSG " + target + " :" + chr(1) + "IMAGE " + url + chr(1) + "\r\n")

  def ctcpMe(self, target, message):
    self.send("PRIVMSG " + target + " :" + chr(1) + "ACTION " + message + chr(1) + "\r\n")

  def stat(self, request, server =""):
    ircCommand = "STATS " + request
    if server != "":
      ircCommand = ircCommand + " " + server

    ircCommand = ircCommand + "\r\n"
    self.send(ircCommand)

  def link(self, server="", maskServer=""):
    ircCommand = "LINKS " + server + " " + maskServer + "\r\n"
    self.send(ircCommand)

  def time(self, server=""):
    ircCommand = "TIME " + server + "\r\n"
    self.send(ircCommand)

  def admin(self, server=""):
    ircCommand = "ADMIN " + server + "\r\n"
    self.send(ircCommand)

  def info(self, server=""):
    ircCommand = "INFO " + server + "\r\n"
    self.send(ircCommand)

  def changeNick(self, newNick):
    ircCommand = "NICK"+" "+ newNick+ "\r\n"
    self.send(ircCommand)

  def __principalLoop(self):
    line            = ""
    gettingDeeperIn = 0
    tooMany         = ""

    while (not gettingDeeperIn and self.running):
      line += self.ircConnection.recv(512)
      tuple = self.getMsgFromBuffer(line)
      line  = tuple[1]

      for token in tuple[0]:
        if (gettingDeeperIn):
          tooMany += token + "\r\n"
        else:
          match = re.search(':\S+\s+(\d\d\d)\s+(\S+)\s+:', token)
          if match is not None and match.group(2) != self.getNick():
            self.setNick(match.group(2))
          if re.search(':(\S+)\s+(376|422) ' + self.getNick(), token) != None:
            gettingDeeperIn = 1
          self.queue.put(("WELCOME", token))

    buffer = tooMany + line

    regexIRC    = re.compile('(?m)^(PING\s+:.*)|^:([^!:]+)!(\S+)\s+([A-Za-z]+)\s+(#?[^: \t]+)?\s*([^:]+)?:?([^\r]*)')
    regexNames  = re.compile(':(\S+)\s+(\S+)\s+(\S+)\s+(.*)')
    regex353    = re.compile("(=|@) (\S+)\s+:(.*)")
    regex332    = re.compile("(\S+)\s+:(.*)")
    regexEnd    = re.compile(":(.*)")

    self.queue.put(("LOGGED", ))

    while self.running:
      tuple  = self.getMsgFromBuffer(buffer)
      liste  = tuple[0]
      buffer = tuple[1]
        
      for ircCommand in liste:
        if irclib_debug: print "*** DEBUG :: " + ircCommand

        match = regexIRC.search(ircCommand)
        if match is not None:

          if match.group(0).find('PING') == 0:
            match = re.search(':(.*)', match.group(1))
            self.queue.put(("PING", match.group(1)))
          else:
            dump, nick, id, ircCommand, source, arg, text = match.groups(5)

            if ircCommand == 'PRIVMSG':
              if source[0] == '#':
                self.queue.put(("MESSAGE", id, nick, source, text))

              else :
                self.queue.put(("PRIVATE", id, nick, text))

            elif ircCommand == 'KICK':
              self.queue.put(("KICK", id, nick, source, arg.strip(), text))

            elif ircCommand == 'JOIN':
              self.queue.put(("JOIN", id, nick, text))

            elif ircCommand == 'QUIT':
              self.queue.put(("QUIT", id, nick, text))

            elif ircCommand == 'PART':
              self.queue.put(("PART", id, nick, source))

            elif ircCommand =='MODE':
              self.queue.put(("CHANGEMODE", id, nick, source, arg))

            elif ircCommand =='NICK':
              self.queue.put(("NICK", id, nick, text))

            elif ircCommand == 'NOTICE':
              if source[0]=="#":
                self.queue.put(("CHANNEL-NOTICE", id, nick, source, text))
              else:
                self.queue.put(("NOTICE", id, nick, source, text))

            elif ircCommand == 'TOPIC':
              self.queue.put(("CHANNEL-TOPIC", id, nick, source, text))

            else:
              self.queue.put(("DEFAULT", ircCommand))

        else:
          match = regexNames.search(ircCommand)
          if match is not None:
            server, ircCommand, nick, argsAndData = match.groups(4)

            if ircCommand == '353':
              if argsAndData[0] == '=' or argsAndData[0] == '@':
                match = regex353.search(argsAndData)
                if match is not None:
                  dump, chan, names = match.groups(3)
                  self.queue.put(("NAMES", chan))
                  for name in string.split(names, " "):
                    self.queue.put(("ADDUSER", chan, name))

            elif ircCommand == '366':
              self.queue.put(("END-NAMES", ))
            elif ircCommand == '332':
              match = regex332.search(argsAndData)
              self.queue.put(("JOIN-TOPIC", match.group(1), match.group(2)))
            elif ircCommand == '301':
              match = regex332.search(argsAndData)
              self.queue.put(("AWAY", match.group(1), match.group(2)))
            elif ircCommand == '305':
              match = regexEnd.search(argsAndData)
              self.queue.put(("SET-AWAY", match.group(1)))
            elif ircCommand == '306':
              match = regexEnd.search(argsAndData)
              self.queue.put(("SET-AWAY", match.group(1)))
            elif ircCommand == '333':
              match = re.search("(\S+)\s+(\S+)\s+(\S+)", argsAndData)
              self.queue.put(("SET-BY", match.group(1), match.group(2), match.group(3)))
            elif ircCommand == '311':
              match = re.search("(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+:(.*)", argsAndData)
              self.queue.put(("WHOIS-IS", match.group(1), match.group(2), match.group(3), match.group(4), match.group(5)))
            elif ircCommand == '319':
              match = regex332.search(argsAndData)
              self.queue.put(("WHOIS-CHANS", match.group(1), match.group(2)))
            elif ircCommand == '312':
              match = re.search("(\S+)\s+(\S+)\s+:(.*)", argsAndData)
              self.queue.put(("WHOIS-SERVER", match.group(1), match.group(2), match.group(3)))
            elif ircCommand == '317':
              match = re.search("(\S+)\s+(\d+)\s+(\d+)\s+:(.*)", argsAndData)
              self.queue.put(("WHOIS-TIME", match.group(1), match.group(2), match.group(3), match.group(4)))
            elif ircCommand == '322':
              match = re.search("(\S+)\s+(\S+)\s+:(.*)", argsAndData)
              self.queue.put(("CHANS-LIST", match.group(1), match.group(2), match.group(3)))
            elif ircCommand == '433':
              match = regex332.search(argsAndData)
              self.queue.put(("NICK-TAKEN", match.group(1), match.group(2)))
            elif ircCommand == 'NOTICE':
              self.queue.put(("SERVER-NOTICE", server, argsAndData[1:]))

      buffer = buffer + self.ircConnection.recv(512)
      if buffer == '':
        break

# end of irclib.py